﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;

namespace Inedo.WindowsServices
{
    /// <summary>
    /// Represents the action the service controller should take on each failure of a service.
    /// A service is considered failed when it terminates without reporting a status of SERVICE_STOPPED to the service controller.
    /// </summary>
    [Undisclosed]
    public sealed class ServiceFailureActions
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="ServiceFailureActions"/> class.
        /// </summary>
        /// <param name="resetPeriod">The reset period in seconds. A value of null indicates that this should never be reset.</param>
        /// <param name="rebootMessage">The message to be broadcast to server users before rebooting in response to the SC_ACTION_REBOOT service controller action.</param>
        /// <param name="command">The command line of the process for the CreateProcess function to execute in response to the SC_ACTION_RUN_COMMAND service controller action.</param>
        /// <param name="actions">The actions to perform.</param>
        /// <exception cref="ArgumentNullException"><paramref name="actions"/> is null.</exception>
        public ServiceFailureActions(int? resetPeriod, string rebootMessage, string command, params ServiceControllerAction[] actions)
        {
            if (actions == null)
                throw new ArgumentNullException(nameof(actions));

            this.ResetPeriod = resetPeriod;
            this.RebootMessage = rebootMessage;
            this.Command = command;
            this.Actions = new List<ServiceControllerAction>(actions).AsReadOnly();
        }
        /// <summary>
        /// Initializes a new instance of the <see cref="ServiceFailureActions"/> class.
        /// </summary>
        /// <param name="resetPeriod">The reset period in seconds. A value of null indicates that this should never be reset.</param>
        /// <param name="rebootMessage">The message to be broadcast to server users before rebooting in response to the SC_ACTION_REBOOT service controller action.</param>
        /// <param name="actions">The actions to perform.</param>
        /// <exception cref="ArgumentNullException"><paramref name="actions"/> is null.</exception>
        public ServiceFailureActions(int? resetPeriod, string rebootMessage, params ServiceControllerAction[] actions)
            : this(resetPeriod, rebootMessage, null, actions)
        {
        }
        /// <summary>
        /// Initializes a new instance of the <see cref="ServiceFailureActions"/> class.
        /// </summary>
        /// <param name="resetPeriod">The reset period in seconds. A value of null indicates that this should never be reset.</param>
        /// <param name="actions">The actions to perform.</param>
        /// <exception cref="ArgumentNullException"><paramref name="actions"/> is null.</exception>
        public ServiceFailureActions(int? resetPeriod, params ServiceControllerAction[] actions)
            : this(resetPeriod, null, null, actions)
        {
        }

        /// <summary>
        /// Gets the time after which to reset the failure count to zero if there are no failures in seconds.
        /// A value of null indicates that this should never be reset.
        /// </summary>
        public int? ResetPeriod { get; }

        /// <summary>
        /// Gets the message to be broadcast to server users before rebooting in response to the SC_ACTION_REBOOT service controller action.
        /// </summary>
        public string RebootMessage { get; }

        /// <summary>
        /// Gets the command line of the process for the CreateProcess function to execute in response to the SC_ACTION_RUN_COMMAND service controller action.
        /// </summary>
        public string Command { get; }

        /// <summary>
        /// Gets the actions to perform.
        /// </summary>
        public ReadOnlyCollection<ServiceControllerAction> Actions { get; }
    }
}
